/*              Flash Pong, by Brad Smith, 1998                  */
/*                                                               */
/*   I learned sources for timers, vga, and other stuff from     */
/*  many different places, so it's hard to give credit to them.  */
/*   I wrote the drawing primitaves myself, so I can take full   */
/*   credit for them, and the rest of this program. If you use   */
/* them be careful with them. I don't check for points offscreen */
/*   or in the wrong order because I didn't need them in this    */
/* program, but that kind of code is easy anyway. So it doesn't  */
/*   matter in that respect. I'm just warning you, that's all.   */
/*                                                               */
/*    Otherwise, this is Flash Pong, by Brad Smith 1998          */
/*                                                               */
/*     Feel free to borrow anything from this source.            */
/*                                                               */
/*  If anyone knows who made the original pong, please tell me.  */
/* Comments or Suggestions: email me at rainwarrior@hotmail.com  */
/* my website http://www.fortunecity.com/skyscraper/oracle/138/  */

#include <dos.h>         /* for keyboard / timer / vga */
#include <stdio.h>       /* for standard input / output */
#include <stdlib.h>      /* for standard library stuff (rand) */
#include <string.h>      /* for memcpy */
#include <sys/nearptr.h> /* for DJGPP memory */
#include <time.h>        /* for uclock timers */
#include <dpmi.h>        /* for keyboard */
#include <go32.h>        /* for keyboard */
#include <pc.h>          /* for keyboard */

typedef unsigned char byte;
byte *vga;   /* vga ram pointer */

char *charset = (char *)0xffa6e; /* text set pointer */

float strtime; /* i like a shared timing variable */

/**/
/* A nice little way to get a random number */
/* REMEMBER TO SEED IT BY CALLING SRANDOM FIRST */
/**/
int genrand(int ran,int add)
{ return (rand() & 65535)/(65536/ran)+add; }
/**/
/* End of that, but here's a quick explanation: */
/* the & is a bitwise operator that only keeps the */
/* values that are in the second number. Since 65535 */
/* has all 16 bits on, we will get a number that is */
/* between 0 and 65535 out of the rand(), which is */
/* normally between 0 and about 2 million. */
/* after that, it's just a simple division, and addition. */
/**/

/**/
/* Start of text stuff */
/**/
void blit_char(int x, int y, int color, unsigned char c)
{ int offset, x2, y2;
  char *work_char;
  unsigned char bit_mask = 0x80;
  work_char = charset + c * 8;
  offset = (y << 8) + (y << 6) + x;
  for(y2=0; y2<8; y2++)
  { bit_mask = 0x80;
    for(x2=0; x2<8; x2++)
    { if((*work_char & bit_mask))
      vga[offset+x2] = color;
    bit_mask = (bit_mask >> 1); }
    offset += 320;
    work_char++;
} }

void blit_string(int x, int y, int color, char *string)
{ int index;
  for(index=0; string[index] != 0; index++)
{ blit_char(x+(index<<3),y,color,string[index]); } }
/**/
/* End of text stuff */
/**/

/**/
/* convert an integer (that is from 0 to 999) to a 3 char string */
/**/
void inttostr(int unwind, char intstr[4])
{   int i;
    int da; /* Since I'm converting to base 10, and I won't */
    int db; /* be using number over 999, i'll only need 3   */
    int dc; /* digits */
    /* the string to change is intstr[4] */
    /* unwind is the number */
    da=0;
    for(i=1;i<11;i++)
    {  if(unwind-100*i < 0){da=i-1; break;} }    /* get the 100's digit */
    db=0;
    for(i=1;i<11;i++)
    {  if(unwind-(da*100+10*i) < 0){db=i-1; break;} }     /* the 10's */
    dc=0;
    for(i=1;i<11;i++)
    {  if(unwind-(da*100+db*10+i) < 0){dc=i-1; break;} }     /* the 1's */
    intstr[0]=da+48;   /* ASCII value of a 1 digit number is 48+n */
    intstr[1]=db+48;
    intstr[2]=dc+48;
    if(da==0){intstr[0]=' ';} /* Get rid of front zeroes */
    if(db==0 && da==0){intstr[1]=' ';}
    intstr[3]='\0';  /* terminate the string */
return;}
/**/
/* no more integer conversion for me */
/**/

/**/
/* keyboard stuff */
/**/
unsigned char raw_key, key_table[128];
int keys_active=0;
	
_go32_dpmi_seginfo old_key_handler,new_key_handler;

void key_handler(void)
{  unsigned char al, ah;
   asm("cli; pusha");
   raw_key = inportb(0x60);
   al = inportb(0x61);
   al |= 0x82;
   outportb(0x61, al);
   al &= 0x7f;
   outportb(0x61, al);

   if(raw_key < 128)
   { if(!key_table[raw_key])
     { key_table[raw_key] = 1;
   keys_active++; } }
   else
   { if(key_table[raw_key-128])
     { key_table[raw_key-128] = 0;
   keys_active--; } }

   outportb(0x20, 0x20);
asm("popa; sti"); }

void key_init(void)
{ int index;
  new_key_handler.pm_offset   = (int)key_handler;
  new_key_handler.pm_selector = _go32_my_cs();
  _go32_dpmi_get_protected_mode_interrupt_vector(0x9, &old_key_handler);
  _go32_dpmi_allocate_iret_wrapper(&new_key_handler);
  _go32_dpmi_set_protected_mode_interrupt_vector(0x9,&new_key_handler);
  for(index=0;index<128;index++)
{ key_table[index] = 0;} }

void key_delete(void)
{ _go32_dpmi_set_protected_mode_interrupt_vector(0x9,&old_key_handler); }
/**/
/* end of keyboard stuff */
/**/

/**/
/* timer stuff */
/**/
uclock_t startuclock;        /* clock variable */

void initftime(void)
{ startuclock=uclock();}    /* setup the clock*/

double ftime(void) /* turn clock ticks into seconds */
{ return (double)(uclock()-startuclock)*1.0/UCLOCKS_PER_SEC; }
/**/
/* end of timer stuff */
/**/

/**/
/* graphics setup stuff */
/**/
int startvga()
{ union REGS IN,OUT;
  IN.h.al = 19;       /* change to mode 13h (19) */
  IN.h.ah = 0x00;
  int86(0x10,&IN,&OUT);
  vga=(byte *)(0xa0000 + __djgpp_conventional_base);
return 1;}

void endvga()
{ union REGS IN,OUT;
  IN.h.al = 0x3;   /* back to text mode, or whatever. */
  IN.h.ah = 0x00;
  int86(0x10,&IN,&OUT);
return;}
/**/
/* end of graphics setup stuff*/
/**/

/**/
/* graphic drawing */
/**/
void plot(int x, int y, byte color)
{ vga[y*320+x]=color; }                 /* set the pixel */

void hline(int x1, int x2, int y, byte color)
{ int i;
  byte xbits[x2-x1+1];
  for(i=0;i<x2-x1+1;i++)
  { xbits[i]=color; }                   /* make a row of the color*/
memcpy(vga+(y*320)+x1,xbits,x2-x1+1); } /* copy the whole row at once */

void vline(int x, int y1, int y2, byte color)
{ int i;
  for(i=y1;i<y2;i++)
{ vga[i*320+x]=color;} }

void box(int x1, int y1, int x2, int y2, byte color)
{ int x,y,i;
  byte xbits[x2-x1+1];
  { for(y=y1;y<y2+1;y++)
    { for(i=0;i<x2-x1+1;i++)
      { xbits[i]=color; }
    memcpy(vga+(y*320)+x1,xbits,x2-x1+1); }
} }

void cls()
{ int i,j;
  for(i=0;i<320;i++)   /* set every pixel to 0 */
  { for(j=0;j<200;j++)
{ vga[j*320+i]=0;} } }
/**/
/* end of graphic drawing*/
/**/

/**/
/* Game routines */
/**/
void drawmenu()
{ cls();
  vline(0,0,199,20);
  vline(319,0,199,20);
  hline(0,319,199,20);
  box(0,0,319,15,20);       /* grey pong background */
  box(10,92,15,112,20);
  box(305,92,310,112,20);
  box(158,100,162,104,20);

  blit_char(120,8,40,'F');
  blit_char(128,8,42,'L');
  blit_char(136,8,44,'A');
  blit_char(144,8,48,'S');
  blit_char(152,8,32,'H');
  blit_char(168,8,42,'P');
  blit_char(176,8,44,'O');
  blit_char(184,8,48,'N');
  blit_char(192,8,32,'G');
  blit_string(120,24,48,"Brad Smith");
  blit_string(8,32,119,"Belly of the Great Plaid Frog Software");
  blit_string(146,40,40,"1998");
  blit_string(68,184,28,"Press [ENTER] to select");
  blit_string(80,64,30,"Computer VS Computer");
  blit_string(80,80,30," Player VS Computer ");
  blit_string(80,96,30,"  Player VS Player  ");
  blit_string(80,112,30,"    Instructions    ");
  blit_string(80,128,30,"        Quit        ");
return;}

void drawinstr()
{ int balcol=32;
  cls();
  vline(0,0,199,20);
  vline(319,0,199,20);
  hline(0,319,199,20);
  box(0,0,319,15,20);    /* grey pong background */
  box(10,92,15,112,20);
  box(305,92,310,112,20);
  box(158,100,162,104,20);

  blit_char(120,8,40,'F');
  blit_char(128,8,42,'L');
  blit_char(136,8,44,'A');
  blit_char(144,8,48,'S');
  blit_char(152,8,32,'H');
  blit_char(168,8,42,'P');
  blit_char(176,8,44,'O');
  blit_char(184,8,48,'N');
  blit_char(192,8,32,'G');
  blit_string(112,24,31,"Instructions");
  blit_string(16,40,23,"For Computer VS Computer mode, just ");
  blit_string(16,48,24," select it, and watch it go.        ");

  blit_string(16,64,25,"For Player VS Computer mode, your   ");
  blit_string(16,72,26," controls are:                      ");
  blit_string(16,80,27,"    A - up / Z - down.              ");

  blit_string(16,96,28,"For Player VS Player mode, the      ");
  blit_string(16,104,29," second player's controls are:      ");
  blit_string(16,112,30,"    K - up / M - down               ");

  blit_string(16,128,31,"  Press escape to end at any time.  ");

  blit_string(16,184,28,"    Press [ENTER] to continue...    ");

  strtime=ftime();               /* 1/10 second wait to make sure */
  while(ftime()-strtime<0.2){}   /* that they had time to release */
  while(!key_table[28] && !key_table[1]) /* the ENTER key */
  { strtime=ftime();
    balcol++;                          /* increment the ball color */
    if(balcol>55){balcol=balcol-24;}
    if(balcol==47){balcol=49;}
    box(114,10,118,14,balcol);         /* draw it around the title */
    box(201,10,205,14,balcol);
    while(ftime()-strtime<0.01){}
  }
  strtime=ftime();
  while(ftime()-strtime<0.2){}
return;}

void rungame(int gamemode)
{ int lpad,rpad,balx,baly; /* pads and ball positions */
  int lpadd,rpadd,bald; /* pads and ball directions */
  int balcol=32; /* ball color */
  char prinscor[4];
  int lscor,rscor;
  int newball=1;
  int commove=0;
  lscor=0;
  rscor=0;
  srand(ftime()); /* seed the random number generator */

  cls();
  vline(0,0,199,8);
  vline(319,0,199,8);
  hline(0,319,199,8);
  hline(0,319,16,8);
  box(0,0,319,15,14);
  box(5,2,34,12,0);
  box(285,2,314,12,0);
  blit_char(120,4,40,'F');
  blit_char(128,4,42,'L');
  blit_char(136,4,43,'A');
  blit_char(144,4,54,'S');
  blit_char(152,4,32,'H');
  blit_char(168,4,42,'P');
  blit_char(176,4,43,'O');
  blit_char(184,4,54,'N');
  blit_char(192,4,32,'G');
  lpad=92;
  rpad=92;          /* paddle should be 20 long */
  balx=160;         /* ball should be 4 x 4   */
  baly=90;
  lpadd=0;
  rpadd=0;
  bald=1;
  srandom(ftime());
  while(!key_table[1])  /* unless ESC is pressed, keep playing */
  { strtime=ftime();
    commove++;
    if(commove>6){commove=0;}
    if(newball==1)
    { balx=160;
      baly=genrand(174,19);
      bald=genrand(3,1);
      newball=0;
      box(5,2,34,12,0);
      box(285,2,314,12,0);
      inttostr(lscor,prinscor);
      blit_string(10,3,28,prinscor);
      inttostr(rscor,prinscor);
      blit_string(290,3,28,prinscor);
    }
    balcol++;                          /* increment the ball color */
    if(balcol>55){balcol=balcol-24;}
    if(balcol==47){balcol=49;}
    if(bald==1){balx++;baly++;}
    if(bald==2){balx--;baly++;}
    if(bald==3){balx++;baly--;}
    if(bald==4){balx--;baly--;}
    box(balx,baly,balx+3,baly+3,balcol); /* draw the ball */
    /* now erase the old ball */
    if(bald==1)
    { hline(balx-1,balx+4,baly-1,0);
      vline(balx-1,baly-1,baly+4,0);}
    if(bald==2)
    { hline(balx-1,balx+3,baly-1,0);
      vline(balx+4,baly-1,baly+4,0);}
    if(bald==3)
    { hline(balx-1,balx+3,baly+4,0);
      vline(balx-1,baly,baly+4,0);}
    if(bald==4)
    { hline(balx+1,balx+4,baly+4,0);
      vline(balx+4,baly,baly+4,0);}
    if(baly==194){bald=bald+2;} /* keep the ball from leaving the area */
    if(baly==18){bald=bald-2;}

    if(balx==16)                     /* check if it hits the paddle */
    {  if(baly>lpad-5 && baly<lpad+21){bald--;} }
    if(balx==300)
    {  if(baly>rpad-5 && baly<rpad+21){bald++;} }

    if(balx==5)           /* check if it scores */
    { newball=1;
      box(balx,baly,balx+3,baly+3,0);
      rscor++;
      if(rscor>999){rscor=0;}
    }
    if(balx==314)
    { newball=1;
      box(balx,baly,balx+3,baly+3,0);
      lscor++;
      if(lscor>999){lscor=0;}
    }
    if(gamemode==1 && commove!=4)   /* if computer is left paddle */
    { if(baly<lpad+8){lpad--; lpadd=1;}
      if(baly>lpad+8){lpad++; lpadd=2;}   /* simple AIish movements */
    }
    if(gamemode<3 && commove!=2)    /* if computer is right paddle */
    { if(baly<rpad+8){rpad--; rpadd=1;}
      if(baly>rpad+8){rpad++; rpadd=2;}
    }

    if(gamemode>1)   /* if player is left paddle */
    { if(key_table[30]){lpad--; lpadd=1;}
      if(key_table[44]){lpad++; lpadd=2;}
    }
    if(gamemode==3)   /* if player is right paddle */
    { if(key_table[37]){rpad--; rpadd=1;}
      if(key_table[50]){rpad++; rpadd=2;}
    }
    if(rpad<18){rpad=18;}
    if(rpad>177){rpad=177;}  /* make sure not to put the paddle */
    if(lpad<18){lpad=18;}    /* out of the playing field        */
    if(lpad>177){lpad=177;}
    box(10,lpad,15,lpad+20,14);   /* draw the paddles */
    box(304,rpad,309,rpad+20,14);
    if(lpadd==1){hline(10,15,lpad+21,0);}
    if(lpadd==2){hline(10,15,lpad-1,0);} /* remove the old paddle edges */
    if(rpadd==1){hline(304,309,rpad+21,0);}
    if(rpadd==2){hline(304,309,rpad-1,0);}
    lpadd=0;
    rpadd=0;
    while(ftime()-strtime<0.01){}
  }
  strtime=ftime();   /* to keep from quitting by accident */
  while(ftime()-strtime<0.2){}
return;}
/**/
/* End of Game routines*/
/**/

void main()
{ int balcol=32;
  int mopt=1;   /* menu option */
  int moptr=0;  /* waiting for key release variable */

  __djgpp_nearptr_enable();   /* Start up nearptr stuff */
  startvga();
  initftime();
  key_init();
  charset += __djgpp_conventional_base; /* start up text */

  drawmenu();
 while(mopt!=5)
 {drawmenu();

  while(!key_table[28] && !key_table[1])
  {
  strtime=ftime();
    balcol++;                          /* increment the ball color */
    if(balcol>55){balcol=balcol-24;}
    if(balcol==47){balcol=49;}
    box(114,10,118,14,balcol);         /* draw it around the title */
    box(201,10,205,14,balcol);
    if(moptr!=0){moptr--;}  /* if skipping presses, decrease number */
    if(mopt==1)
    { box(66,66,70,70,balcol); box(249,66,253,70,balcol);}
    if(mopt==2)             /* draw it around the current selection */
    { box(66,82,70,86,balcol); box(249,82,253,86,balcol);}
    if(mopt==3)
    { box(66,98,70,102,balcol); box(249,98,253,102,balcol);}
    if(mopt==4)
    { box(66,114,70,118,balcol); box(249,114,253,118,balcol);}
    if(mopt==5)
    { box(66,130,70,134,balcol); box(249,130,253,134,balcol);}
    if(key_table[72] && !moptr)
    { mopt--;
      moptr=20;         /* skip 20 cycles to prevent multiple presses */
      if(mopt<1){mopt=5;}
      box(66,66,70,70,0); box(249,66,253,70,0);
      box(66,82,70,86,0); box(249,82,253,86,0);
      box(66,98,70,102,0); box(249,98,253,102,0);
      box(66,114,70,118,0); box(249,114,253,118,0);
      box(66,130,70,134,0); box(249,130,253,134,0);
    }
    if(key_table[80] && !moptr)
    { mopt++;
      moptr=20;
      if(mopt>5){mopt=1;}
      box(66,66,70,70,0); box(249,66,253,70,0);
      box(66,82,70,86,0); box(249,82,253,86,0);
      box(66,98,70,102,0); box(249,98,253,102,0);
      box(66,114,70,118,0); box(249,114,253,118,0);
      box(66,130,70,134,0); box(249,130,253,134,0);
    }
    while(ftime()-strtime<0.01){}
  }
  if(key_table[1]){ mopt=5;}
  if(mopt<4){ rungame(mopt);} /* options 1 2 3 are all the game */
  if(mopt==4){ drawinstr();}
 }

  endvga();          /* back to text mode */
  key_delete();      /* get rid of keypress checking */
  __djgpp_nearptr_disable();  /* make sure to put DPMI back */
  printf("Goodbye!\n");
return;}
